#-
*    Copyright (C) 2013 Piotr Kotko
*
*    This file is part of OGIME
*
*    OGIME is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.



*====================================== user defined ============================================
*--- define the number of simplification cycles
#define SIMPCYCLES "40"
*================================================================================================

#setrandom 12122121

*--- declare variables and restore some settings
#include dimensions.frm
#include declarations.frm
#include `OUTPUT_DIR'/`OUTPUT_FILE'.inc
#include globalvars.frm
#include print.frm
#if `FORMSTAT' == 0
off stats;
#procedure UserRelations
#endprocedure
#endif

*--- order in the strong coupling (at tree-level)
#define GORD "{`NFS'-2}"

#write "\\n"
#call info(Checking gauge invariance of `OUTPUT_FILE'...)

*--- load the first color amplitude
load `OUTPUT_FILE'.sav;

drop;
ndrop amp1;

L amp = amp1;

multiply replace_(dum_,fun);

#call print
#$nden = 1; *>> will be needed below
b Den;
.sort: loading;


*-----------------------------------------------------------------------------------------------


*--- take the expression to the common denominator (and reject the denominator)
*--- by denominator we mean here function Den
#call info(Taking expressions to a common denominator...)
keep brackets;
repeat;
if ( match(once Den(ARG?) ) );
once Den(ARG?) = Den($nden,ARG);
$nden = $nden+1;
endif;
endrepeat;
#call print
moduleoption,maximum,$nden; 
.sort: Common den A;

#do k=1,{`$nden'-1}
id Den(`k',ARG?$den`k') = Den(`k',ARG);
#enddo
.sort: Common den B;
cf Num;
#if `$nden' > 1
multiply Num(1,`$den1')
#do k=2,{`$nden'-1}
          *Num(`k',`$den`k'')
#enddo
                          ;
#endif

#call print
b Num,Den;
.sort: Common den C;
*-----------------------------------------------------------------------------------------------


*--- cancel numerators and denominators
keep brackets;
repeat id Num(ARG1?,?a1)*Den(ARG1?,?a2) = 1;
id Num(?a,ARG1?) = Num(ARG1);
.sort: Cancel;

*--- replace polarization vector by momentum
multiply replace_(e`REPL_NO',k`REPL_NO',ec`REPL_NO',k`REPL_NO');
#call print
.sort: Replace;
*-----------------------------------------------------------------------------------------------

#call info(Pre-simplifying expressions...)

*--- simplify
#include simplifymom.frm
#include simplifyinv.frm
.sort:Simplify A;

#call SimplifyMom(fun);
#call SimplifyInv(fun);

#call info(Expanding numerators...)

id Num(ARG?) = ARG;
id fun(ARG?) = ARG;

.sort:Simplify B;
#$nterms = termsin_(amp);
#call info(Initial number of terms: `$nterms')
#call info(Simplifying...)
#do l=1,`SIMPCYCLES'
#$nterms = termsin_(amp);
#call SimplifyMomExpr(amp)
#call GenerateInvGene
#call SimplifyInvExpr(amp)
#$nterms1 = termsin_(amp);
#call info(Number of terms left: `$nterms1')
#if `$nterms1' == `$nterms'
#redefine SIMPERTURB "{`SIMPERTURB'-10}"
#else
#redefine SIMPERTURB "0"
#endif
#if `$nterms1' == 0
#redefine l "{`SIMPCYCLES'}"
#endif
#enddo


print;

.end
