*    Copyright (C) 2013 Piotr Kotko
*
*    This file is part of OGIME
*
*    OGIME is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.



*-------------------------------------------------------------------------------------------------------------
*--------------------------------- prepare tables of relations between invariants ----------------------------
*-------------------------------------------------------------------------------------------------------------

*--- we will define tables containing all the relations between invariants to be used in simplifications




*--- define local variable with the number of final partons
#define N "`NFS'"

*--- the maximal number of invariants
#define MAXNINV "{`N'!/(2*(`N'-2)!)+`N'}"


hide;
.sort:Invariants;
*-------------------------------------------------------------------------------------------------------------


*--- prepare the table of momenta permutations first
auto v vtm;
auto s arg, ii;
s xcf, ycf, Two, Minus;
i nu;
cf ftmp, gtmp;

table moms(1:{`N'!},1:`N');

L tmp = gtmp(
#do l=1,{`N'-1}
             ftmp(k`l'(nu))*
#enddo
                       ftmp(k`N'(nu)));

argument gtmp;
shuffle, ftmp;
endargument;

splitarg gtmp;

$powy = 1;

argument gtmp;
$powx = 1;
argument ftmp;
multiply xcf^$powx;
$powx = $powx + 1;
endargument;

multiply ycf^$powy;
$powy = $powy +1;

endargument;

repeat
id gtmp(arg1?,arg2?,?a) = gtmp(arg1+arg2,?a);

id gtmp(arg?) = arg;

repeat
id ftmp(arg1?,arg2?,?a) = ftmp(arg1+arg2,?a);

id ftmp(arg?) = arg;

b xcf, ycf;

.sort:Invariants;

*--- fill the table with momenta permutations
fillexpression moms = tmp(ycf,xcf);

.sort:Invariants;
*-------------------------------------------------------------------------------------------------------------


*--- prepare and define relations between invariants
drop tmp;
#define NINVTABS "0"

#do p=1,`N'

L ident = k1+...+k`N';

.sort:Invariants;

#redefine NINVTABS "{`NINVTABS'+1}"

*--- this will store a relation between invariants, the sum of all components will be zero
table zerofill invtab`NINVTABS'(1:`MAXNINV');

id vtm? = k`p'.vtm;

#call kiki

$pow = 1;

.sort:Invariants;

multiply xcf^$pow;

$pow = $pow + 1;

b xcf;

.sort:Invariants;

*--- fill the table with invariants
fillexpression invtab`NINVTABS' = ident(xcf);

#define INVDIM`NINVTABS' "{`$pow'-1}"

#enddo

.sort:Invariants;


#do j=1,{`N'!}
#do p=1,'N'


#redefine NINVTABS "{`NINVTABS'+1}"

table zerofill invtab`NINVTABS'(1:`MAXNINV');

L ident = 1;

$v1 = sum_(ii,1,`p',moms(`j',ii))^2;
$v2 = sum_(ii,{`p'+1},`N',moms(`j',ii))^2;

multiply ($v1 - ($v2) );

if ( coefficient == -2 );
multiply (-Minus*Two/2);
elseif ( coefficient == 2 );
multiply Two/2;
elseif ( coefficient == -1 );
multiply (-Minus);
endif;

contract;

.sort:Invariants;

#call kiki

$pow = 1;

.sort:Invariants;

multiply xcf^$pow;

$pow = $pow + 1;

b xcf;

.sort:Invariants;

#$dim = $pow;

fillexpression invtab`NINVTABS' = ident(xcf);

#define INVDIM`NINVTABS' "{`$pow'-1}"

.sort:Invariants;

#enddo
#enddo
*-------------------------------------------------------------------------------------------------------------

*--- print out the tables with invariants
#if `PRINT' == 1
#write "Printing the table with relations between invariants..."
#do i=1,`NINVTABS'
printtable invtab`i';
#enddo
.sort:Invariants;
#endif

unhide;
drop ident;
#undefine N

.sort:Invariants;

