#-
*    Copyright (C) 2013 Piotr Kotko
*
*    This file is part of OGIME
*
*    OGIME is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.


#setrandom

*========================================= matrix element ======================================

*--- declare variables
#include dimensions.frm
#include declarations.frm
#include globalvars.frm
#include print.frm
#if `FORMSTAT' == 0
off stats;
#endif
#if `SAVEREC' == 1
on checkpoint 5m;
#endif

*-----------------------------------------------------------------------------------------------

*--- print version
#write "\\n"
#call info(Starting OGIME v`VERSION'...)

*--- print user input file that we are using
#call info(Using \\"`INPUT_FILE'.in\\" user input file)


*--- read the matrix element
L ME = 
#include `INPUT_DIR'/`INPUT_FILE'.in
;

.sort: ME A;
*-----------------------------------------------------------------------------------------------

#if `FORMSTAT' == 1
on stats;
#endif

*--- determine the number of final states and off-shell gluons
$nonshell1 = count_([<g|],1);
$nonshell2 = count_([|g>],1);
$nonshell = $nonshell1 + $nonshell2;
$noffshell = count_(W,1);
$nfs = $nonshell + $noffshell;
.sort: ME B;

*--- number of final states
#define NFS "`$nfs'"

*--- number of off-shell legs
#define NOFF "`$noffshell'"

*--- order in the strong coupling (at tree-level)
#define GORD "{`NFS'-2}"
*-----------------------------------------------------------------------------------------------


#call info(The number of final states: `NFS')
#call info(The number of off-shell gluons: `NOFF')



*======================================= calculation part ======================================

*--- read necessary modules
#include interactions.frm
#include gaugelinks.frm
*-----------------------------------------------------------------------------------------------


*--- expand the exponents in gauge fields
#call interaction
#call Wilsonlines

b g;
#call print
.sort: expand exponents;
*-----------------------------------------------------------------------------------------------


*--- retain only the required power in g, and call the coupling g`GORD'

collect fun;
once g^`GORD' = g`GORD';
if ( ( count(g`GORD',1) != 1 ) || ( count(g,1) > 0 ) ) discard;
id fun(ARG?) = ARG;

ab A;
#call print
.sort: order in g;
*-----------------------------------------------------------------------------------------------


*--- "time order" the fields
collect T;
.sort: time order A;
cf [:];  *>> This is trick: make [:] commuting function for a while
b f, t;
#call print
.sort: time order B;
*-----------------------------------------------------------------------------------------------


*--- utilize Wick theorem
#call info(Making Wick contractions...)
#include wick_theorem.frm
*-----------------------------------------------------------------------------------------------


*--- utilize cration/anihilation operators
#call info(Utilizing creation/anihilation operators...)
#include creation_anihilation.frm
*-----------------------------------------------------------------------------------------------


*--- calculate gluon and eikonal propagators
#call info(Calculating eikonal propagators...)
#include propagators.frm
*-----------------------------------------------------------------------------------------------


*--- calculate and utilize momnetum conservation delta functions
#call info(Utilizing momentum conservation delta functions...)
#include deltas.frm
*-----------------------------------------------------------------------------------------------




*===================================== simplification part =====================================


*--- include simplification modules
#call info(Simplifying...)
#include simplifymom.frm
#include simplifyinv.frm
*-----------------------------------------------------------------------------------------------

*``````````````````````````````````````````````````````````````````````
* normalize denominators
#procedure NormDen(foo)
if ( match(`foo'(?a)) );
id `foo'(?a) = `foo'(`foo'(?a));
argument `foo';
normalize `foo';
endargument;
id Den(ARG?) = ARG^-1;
id Den(?a)^-1 = Den(?a);
endif;
.sort:NormDen;
#endprocedure
*``````````````````````````````````````````````````````````````````````


*--- simplify propagators
v K;
id S1(v?) = S1(v.K); *>> K is an auxiliary vector
.sort: Simplify propagators;
#call SimplifyMom(Den)
#call SimplifyMom(S1)
argument S1;
id v?.K = v; *>> remove K
endargument;
id S1(v?) = Den(v.v);
#call Denden;
*-----------------------------------------------------------------------------------------------


*--- transform D functions (numerator of gluon propagator) into a single-argument function (needed by SimplifyMom)
#do j=1,`$NCMAX'
once D(v1?,alph?,bet?) = D`j'A(v1)*D`j'B(alph,bet);
#enddo
.sort: D->Dj;
#do j=1,`$NCMAX'
#call SimplifyMom(D`j'A)
#enddo
*--- compose back D 
#do j=1,`$NCMAX'
id D`j'A(?a1)*D`j'B(?a2) = D(?a1,?a2);
#enddo
.sort: Dj->D;
*-----------------------------------------------------------------------------------------------


*--- replace functional polarization vectors by true vectors, i.e. ef(k'i',alph) -> e'i'(alph)

#do m = 1, `NFS'
id ecf(-k`m',alph?) = ec`m'(alph);
id ef(k`m',alph?) = e`m'(alph);
#enddo
b D;
#call print
.sort: ef->e;
*-----------------------------------------------------------------------------------------------


*--- insert numerator of gluon propagator in Feynman gauge (uncomment to have axial gauge with gauge vector n)
keep brackets;

id D(v1?,mu?,nu?) = (d_(mu,nu)
*                          -(n(mu)*v1(nu)+n(nu)*v1(mu))*Den(n.v1)
*                          + n.n*v1(mu)*v1(nu)*Den(n.v1)^2
                          );
#call kiki
#call eiki
#call nki
#call print;
.sort: Propagator numerator;

#call Denden
#call SimplifyMom(Den)
#call NormDen(Den)
#call Denden
*-----------------------------------------------------------------------------------------------


*--- simplify the whole expression

id den(ARG?) = ARG^-1;
b g`GORD', f, t, Den, den;
.sort: Simplify whole;

collect fun;
.sort: Simplify whole;

#do u=1,`TOTSIMPDEPTH'
#do w=1,`MOMSIMPDEPTH'
#call SimplifyMom(fun)
#enddo
#if `USEINVSIMP' == 1
#call SimplifyInv(fun)
#endif
id fun(0) = 0;
#call print
.sort: Simplify whole;
#enddo
*-----------------------------------------------------------------------------------------------



*=================================== color calculation part ====================================


*--- rearrange color coefficients to become traces of permutations of the generators
#call info(Calculate color amplitudes...)
#include color.frm
*-----------------------------------------------------------------------------------------------

*--- write the preprocessor settings that might be needed in further modules
#write <`OUTPUT_DIR'/`INPUT_FILE'.inc> "#define NFS \"`NFS'\""
#write <`OUTPUT_DIR'/`INPUT_FILE'.inc> "#define NOFF \"`NOFF'\""
#do k=1,`NOFF'
#write <`OUTPUT_DIR'/`INPUT_FILE'.inc> "#$wlinemom`k' = `$wlinemom`k'';"
#write <`OUTPUT_DIR'/`INPUT_FILE'.inc> "#$wlinedir`k' = `$wlinedir`k'';"
#enddo
*-----------------------------------------------------------------------------------------------


*--- collect color (dual) amplitudes (coefficient of the traces) in separate expressions
*--- the name will be amp`i' with `i' corresponding to subsequent color orderings
*--- the first amplitude is with order 1,...,NFS and in practice this one is enough

b Tr, g`GORD';
#$numb = 1;
.sort: Dual amps A;

keep brackets;
once Tr(?a) = Tr($numb,?a);
$numb = $numb + 1;
b Tr, g`GORD';
moduleoption noparallel; 
.sort: Dual amps B;
keep brackets;
#do m = 1, {(`NFS'-1)!}
once Tr(`m',?a$ord`m') = Tr(?a);
b+ Tr;
moduleoption noparallel;
.sort: Dual amps C;
G amp`m' = Tr(`$ord`m'')*ME[Tr(`$ord`m'')];
*--- replace function fun by the function dum_ which is not printed
multiply replace_(fun,dum_);
.sort: Dual amps D;
*--- write the result in a text file
#write <`OUTPUT_DIR'/`INPUT_FILE'.txt> "amp`m' = %e", amp`m'
#enddo
drop ME;
#call print
.store
*--- save the result for further processing
save `INPUT_FILE'.sav;
#call info(Done.)
.end

