*    Copyright (C) 2013 Piotr Kotko
*
*    This file is part of OGIME
*
*    OGIME is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.




*-------------------------------------------------------------------------------------------------------------
*---------------------------- simplification using relation between invariants conservation ------------------
*-------------------------------------------------------------------------------------------------------------


*--- create tables with relations between invariants
#include invariants.frm



*--- determine the length of the "gene" for simplification using invariants depending on the user choice
#define INVGENE "{0}"
#if `SIMPDEPTH' == half
#define INVSIMPSIZE "{`NINVTABS'/2}"
#elseif `SIMPDEPTH' == full
#define INVSIMPSIZE "`NINVTABS'"
#else
#message ERROR! Wrong SIMPDEPTH, SIMPDEPTH = `SIMPDEPTH'.
#terminate
#endif



*=============================================================================================================
#procedure GenerateInvGene
*--- prepare a random "gene" with the numbers indicating which relation between 
*--- invariants (i.e. which `i' in invtab`i') is attempted

hide;
.sort: GenerateInvGene;
cf invgenef;
L invgene = ranperm_(invgenef,1,...,`NINVTABS');
transform,invgenef,replace(`INVSIMPSIZE',last) = ();
id invgenef(?a$invgene) = invgenef(?a);
.sort: GenerateInvGene;
drop invgene;
unhide;
#redefine INVGENE "{`$invgene'}"
.sort: GenerateInvGene;
#endprocedure
*=============================================================================================================



*=============================================================================================================
#procedure SimplifyInv(fun)
*--- simplify function fun using tables of relations between invariants
*--- the invariant to be replaced is chosen according to the generated (random) gene
*--- all the replacements are tried and if the number of terms is smaller the new expression is stored

#call GenerateInvGene

#do u = `INVGENE'
#do w = 1,`INVDIM`u''


b `fun';
.sort:SimplifyInv(`fun') A;

keep brackets;

id `fun'(ARG?) = `fun'(ARG,ARG*(-Minus)*Two/2);

argument 'fun',2;

id invtab`u'(`w') = (-sum_(ii,1,`MAXNINV',invtab`u'(ii)) + invtab`u'(`w'));

#call eiki
#call kiki

id Two = 2;
id Minus = -1;

endargument;

id `fun'(ARG1?,ARG2?) = `fun'(ARG1,nterms_(ARG1),ARG2,nterms_(ARG2));

repeat;
$Ant = 0;
$Bnt = 0;

once 'fun'(ARG1?,ARG2?$Ant,ARG3?,ARG4?$Bnt) = `fun'(ARG1,ARG3);

$score = $Ant - $Bnt;

if ($score >= 0);

#if `VERBOSE' > 1
if ($score > 0);
print "SimplifyInv(`fun'): %$ term(s) simplified out.", $score;
endif;
#endif

id `fun'(ARG1?,ARG2?) = `fun'(ARG2);

else;

id `fun'(ARG1?,ARG2?) = `fun'(ARG1);

endif;

endrepeat;

moduleoption, local, $Ant, $Bnt, $score;
.sort:SimplifyInv(`fun') B;
#enddo
#enddo

#endprocedure
*=============================================================================================================



*=============================================================================================================
#procedure SimplifyInvExpr(fun)
*--- using this procedure one may simplify whole expression (not just a function argument) what is useful
*--- for very large expressions

#call GenerateInvGene

#do u = `INVGENE'
#do w = 1,`INVDIM`u''

skip;

#$Ant = termsin_(`fun');
#$tab = invtab`u'(`w');

#inside $tab;
 id v1?.v2? = fun(v1.v2);
 id fun(ARG?$tabb) = ARG;
#endinside


.sort:SimplifyInvExpr(`fun') A;

hide `fun';

L tmp = `fun';

multiply (-Minus)*Two/2;

b `$tabb', Two, Minus;
.sort:SimplifyInvExpr(`fun') B;

keep brackets;

id `$tab' = (-sum_(ii,1,`MAXNINV',invtab`u'(ii)) + invtab`u'(`w'));

#call eiki
#call kiki

id Two = 2;
id Minus = -1;

.sort:SimplifyInvExpr(`fun') C;

skip;
#$Bnt = termsin_(tmp);
#$score = $Ant - $Bnt;

.sort:SimplifyInvExpr(`fun') D;

unhide `fun';

.sort:SimplifyInvExpr(`fun') E;
skip;
*--- SIMPERTURB is perturbation if simplification stucks (defined in simplifymom.frm)
#if ( `$score'>= `SIMPERTURB' )
#if `VERBOSE' > 1
#if `$score' > 0
#redefine SIMPERTURB "0"
#write "SimplifyInvExpr(`fun'): `$score' term(s) simplified out."
#elseif `$score' < 0
#write "SimplifyInvExpr(`fun'): `$score' term(s) more appeared."
#endif
#endif

#exchange `fun',tmp

#endif

drop tmp;

.sort:InvSimplify(`fun') D;

#enddo
#enddo

unhide;

.sort:InvSimplify(`fun') E;

#endprocedure
*=============================================================================================================

