*    Copyright (C) 2013 Piotr Kotko
*
*    This file is part of OGIME
*
*    OGIME is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.





*-------------------------------------------------------------------------------------------------------------
*--------------------------------- simplification using momentum conservation --------------------------------
*-------------------------------------------------------------------------------------------------------------


*=============================================================================================================
#procedure nki
*--- utilize condition that the gauge link "momentum" is transverse to the direction of the gauge link
#do k=1,`NOFF'
id `$wlinemom`k''.`$wlinedir`k'' = 0;
#enddo
*--- call user restrictions
#call UserRelations
#endprocedure


*=============================================================================================================
#procedure eiki
*--- utilize transversality of polarization vectors
#do j={`NOFF'+1},`NFS'
id e'j'.k'j' = 0;
id ec'j'.k'j' = 0;
#enddo
*--- utilize gauge link direction transversality to an off-shell momentum
#call nki 
#endprocedure
*=============================================================================================================


*=============================================================================================================
#procedure kiki
*--- utilize on-shellnes
#do j={`NOFF'+1},`NFS'
id k'j'.k'j' = 0;
#enddo
#endprocedure
*=============================================================================================================



*=============================================================================================================
#procedure Denden
*--- replace Den by den if there is just one term
id Den(?a) = Den(nterms_(?a),?a);
id Den(1,?a) = den(?a);
id Den(ARG?,?a) = Den(?a);
.sort: Den->den;
#endprocedure
*=============================================================================================================




*=============================================================================================================
*--- the gene is a set with momenta to be replaced using the momentum conservation
#define GENE "{0}"

#procedure GenerateGene
*--- prepare a random gene
hide;
.sort: GenerateGene A;
auto cf genef;
#do j=1,`MOMSIMPDEPTH'
L gene`j' = ranperm_(genef`j',1,...,`NFS');
#enddo
#do j=1,`MOMSIMPDEPTH'
id genef`j'(?a$gene`j') = genef`j'(?a);
#enddo
.sort: GenerateGene B;
#do j=1,`MOMSIMPDEPTH'
drop gene`j';
#enddo
unhide;
#define GENETMP "`$gene1'"
#do i=2,`MOMSIMPDEPTH'
#redefine GENETMP "`GENETMP',`$gene`i''"
#enddo
#redefine GENE "{`GENETMP'}"
.sort: GenerateGene C;
#endprocedure
*=============================================================================================================



*=============================================================================================================
#define MomTables "0"
#procedure SimplifyMom(fun)
*--- simplify the function fun using momentum conservation
*--- the momentum to be replaced is chosen according to the generated (random) gene
*--- all the replacements are tried and if the number of terms is smaller the new expression is stored

#if (`MomTables' == 0)
#redefine MomTables "1"
#call GenerateGene
*--- create a table for SimplifyMom
table check, strict, tabmom(1:`NFS');
fill tabmom(1)=k1,...,k`NFS';
#endif

#do u = `GENE'

$vecA = tabmom(`u');
$vecB = (-sum_(ii,1,`NFS',tabmom(ii))+tabmom(`u'));

b `fun';

.sort:SimplifyMom(`fun') A;

keep brackets;

id `fun'(ARG?) = `fun'(ARG,ARG);

argument `fun',2;
 id `$vecA' = `$vecB';
#call eiki
#call kiki
endargument;

id `fun'(ARG1?,ARG2?) = `fun'(ARG1,nterms_(ARG1),ARG2,nterms_(ARG2));

repeat;
 $Ant = 0;
 $Bnt = 0;
 once `fun'(ARG1?,ARG2?$Ant,ARG3?,ARG4?$Bnt) = `fun'(ARG1,ARG3);
 $score = $Ant - $Bnt;
 if ( $score >= 0 );
#if `VERBOSE' > 1
if ($score > 0);
  print "SimplifyMom(`fun'): %$ term(s) simplified out.", $score;
endif;
#endif
  id `fun'(ARG1?,ARG2?) = `fun'(ARG2);
 else;
  id `fun'(ARG1?,ARG2?) = `fun'(ARG1);
 endif;
endrepeat;

moduleoption, local, $Ant, $Bnt, $vecA, $vecB;
.sort:SimplifyMom(`fun') B;
#enddo

#endprocedure
*=============================================================================================================




*=============================================================================================================
#procedure SimplifyMomL(fun)
*--- version of SimplifyMom which can deal witm bigger expressions
cf nargs;

#if (`MomTables' == 0)
#redefine MomTables "1"
#call GenerateGene
#message "GENE = `GENE'"
*--- create a table for SimplifyMom
table check, strict, tabmom(1:`NFS');
fill tabmom(1)=k1,...,k`NFS';
#endif

#$cnt = 1;

#do u = `GENE'

#$vecA = tabmom(`u');
#$vecB = (-sum_(ii,1,`NFS',tabmom(ii))+tabmom(`u'));

b `fun';

.sort:MomSimplify(`fun') A;

keep brackets;

id `fun'(ARG?) = `fun'($cnt,nterms_(ARG),ARG);

$cnt = $cnt + 1;

#do m=1,{`$cnt'-1}

b `fun';

.sort: new;

keep brackets;

id `fun'(`m',ARGNT?$Ant,ARG?$match) = `fun'(`m',ARG);

.sort:new B;
hide;

L tmp = `$match';

id `$vecA' = `$vecB';

#call eiki
#call kiki

.sort
#$Bnt = nterms_(tmp);
#$score = $Ant - $Bnt;
.sort

drop tmp;
unhide;


#if ( `$score' > 0 )
 print "MomSimplify(`fun'): rule u=`u' applied; %$ terms removed.", $score;
  id `fun'(`m',?a) = `fun'(`m',tmp);
#else
  id `fun'(`m',?a) = `fun'(`m',`$match');
#endif

moduleoption, local, $Ant, $Bnt, $vecA, $vecB;
.sort:MomSimplify(`fun') B;

#enddo

#enddo

id `fun'(ARG?,?a) = `fun'(?a);
.sort:MomSimplify(`fun') C;
#endprocedure
*=============================================================================================================

*--- perturbation for SimplifyMomExpr and SimplifyInvExpr if they stuck; can be set to negative value outside
*--- simplification procedures
#define SIMPERTURB "0"


*=============================================================================================================
#procedure SimplifyMomExpr(fun)

#call GenerateGene

#if (`MomTables' == 0)
#redefine MomTables "1"
*--- create a table for SimplifyMomExpr
table check, strict, tabmom(1:`NFS');
fill tabmom(1)=k1,...,k`NFS';
#endif


#do u = `GENE'

skip;

#$vecA = tabmom(`u');
#$vecB = (-sum_(ii,1,`NFS',tabmom(ii))+tabmom(`u'));

#$Ant = termsin_(`fun');

.sort:SimplifyMomExpr(`fun') A;

hide `fun';

L tmp = `fun';

.sort:SimplifyMomExpr(`fun') B;

id `$vecA' = `$vecB';

#call eiki
#call kiki

**$control = 1;
**else;
**$control = 0;
**endif;

moduleoption, local, $vecA, $vecB;
.sort:SimplifyMomExpr(`fun') B;
skip;
#$Bnt = termsin_(tmp);
#$score = $Ant - $Bnt;
.sort:SimplifyMomExpr(`fun') C;

unhide `fun';

.sort:SimplifyMomExpr(`fun') D;

*--- SIMPERTURB is perturbation if simplification stucks
#if ( `$score'>= `SIMPERTURB' )

#if `VERBOSE' > 1
#if `$score' > 0
#write "SimplifyMomExpr(`fun'): `$score' term(s) simplified out."
#endif
#endif

#exchange `fun',tmp

#endif

drop tmp;

.sort:SimplifyMomExpr(`fun') E;

#enddo

unhide;
.sort:SimplifyMomExpr(`fun') F;

#endprocedure
*```````````````````````````````````````````````````````````````````````


