*    Copyright (C) 2013 Piotr Kotko
*
*    This file is part of OGIME
*
*    OGIME is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.



*-------------------------------------------------------------------------------------------------------------
*---------------------------------------- apply Wick theorem -------------------------------------------------
*-------------------------------------------------------------------------------------------------------------




*--- count the number of fields and number of necessary contractions
cf Tn; *>> stores the number of contractions
cf Ta; *>> stores the number of fields
argument T;
$tc = count_(A,1);
id A(?a) = [:](A1(?a));
endargument;
*>> the number of contractions giving nonzero result (at tree level) is ("number of fields"-NFS+NOFF)/2, put this number to Tn
multiply Ta($tc)*Tn(integer_(($tc-`NFS'+`NOFF')/2));
normalize T;
#call print
moduleoption local, $tc;
.sort: Wick A1;

*--- determine the maximal numbers of contractions and field operators

#$NAMAX = 0;
#$NCMAX = 0;

id Ta(ARG1?$natmp)*Tn(ARG2?$nctmp) = Ta(ARG1)*Tn(ARG2);
if ( $natmp > $NAMAX ) $NAMAX = $natmp;
if ( $nctmp > $NCMAX ) $NCMAX = $nctmp;
moduleoption local, $natmp, $nctmp;
#$NATERMMAX = 0; *>> the maximal number the terms with the same number of fields (prepare for the modules below)
.sort: Wick A2;
*-------------------------------------------------------------------------------------------------------------


*--- define new expressions for terms with different number of field operators


cf Taa; *>> stores the number of field products with the same number of fields
*>> here `nf' is number of fields
#do nf=1,`$NAMAX'
#$NATERM = 0; *>> numbers the terms with the same number of fields
if ( match( Ta(`nf') ) );
$NATERM = $NATERM + 1;
multiply Taa($NATERM);
endif; 
if ( $NATERM > $NATERMMAX );
$NATERMMAX = $NATERM;
endif;
moduleoption noparallel;
.sort: Wick B1a;
#enddo
#call print
.sort: Wick B1b;

*--- store field products in dolar variables $PROD`A'`B'
*--- where A = <number of fields>, B = <number of subsequent term with given number of fields>
*--- the number of fields is stored in $NF`A'`B'
*--- the number of contraction is stored in $NC`A'`B'
cf Tb, Tbb, Tc; *>> auxiliary
#$LISTC = Tbb;
#do nf=1,`$NAMAX'
#do nfa=1,`$NATERMMAX'
if ( match( Ta(`nf')*Taa(`nfa')*T(ARG?$PROD`nf'`nfa')));
id Ta(ARG?$NF`nf'`nfa') = Ta(ARG);
id Ta(`nf')*Taa(`nfa') = Ta(`nf'`nfa');
$LISTC = $LISTC*Tb(`nf'`nfa'); *>> will define $LIST with configurations `A'`B' below
id Tn(ARG?$NC`nf'`nfa') = Tn(ARG);
endif; 
#enddo
#enddo
*--- now Ta stores `A'`B'
#call print
.sort: Wick B2;

*--- put all possible `A'`B' into $LIST
skip;
#inside $LISTC
repeat once Tbb(?a)*Tb(ARG?) = Tbb(?a,ARG);
if (match(Tbb(?a$LIST)));
else;
exit "ERROR: module <Wick theorem C>";
endif;
#endinside
moduleoption noparallel;
.sort: Wick theorem C;

*--- produce new expressions prod`A'`B'
#redefine LIST "{`$LIST'}"
*>> the trick below is because preprocessor turns single lement list {n} into n
#if ( `LIST' == `$LIST' )
#redefine LIST "`$LIST',`$LIST'"
#endif
#do nf = `LIST'
L prod`nf' = (`$PROD`nf'')*Tn($NC`nf');
#enddo
#call print;
.sort: Wick theorem D;

hide;
.sort: Wick theorem E;
*-------------------------------------------------------------------------------------------------------------


*=================================== CONTRACT PROCEDURE =========================================
*--- makes Wick contraction
*--- it tries a random permutations until the correct number of terms is produced
*--- this is much faster than trying all combinations
#procedure contract
if ( ( match( funX(0) ) ) == 0 );
repeat;
argument fun;
id [:](A1(x?,alph?,jj?,ARG1?),A1(y?,bet?,kk?,ARG2?),?b) = S(x,y,alph,bet,ARG1,ARG2)*d_(jj,kk)*deltap_(ARG1,ARG2)*[::](?b)+ranperm_([:],A1(x,alph,jj,ARG1),A1(y,bet,kk,ARG2),?b);
symmetrize S 1,2;
symmetrize S 3,4;
symmetrize S 5,6;
symmetrize [::];
endargument;
id fun(?a) = fun(?a)*fun1(nterms_(?a)-1); 

if ( match( fun1(ARG?)*funX(ARG?) ) );
goto exit1;
endif;
id fun1(?a)=1;
endrepeat;
label exit1;
endif;
id fun(ARG?) = ARG;

*--- remove loops if remain
#do i=1,`NFS'
id 
#do j=1,{`i'-1}
     S(?a`j',ARG`j'?,ARG{`j'+1}?)*
#enddo
     S(?a`i',ARG1?,ARG`i'?) = 0;  
#enddo

id fun1(?a)=1;
id funX(?a)=1;
id [:](?a) = 0;
id [::](?a) = [:](?a);
repeat id [:](?a1,A1(?b1),ARG?) = [:](?a1,A1(?b1))*[:](ARG);
#endprocedure
*================================================================================================


*================================================================================================
#procedure CountWickTerms
*--- count the number of terms with Wick contraction
*--- this number will be stored in funX and used to stop the generation of Wick contractions
id [:](A1(?a1,ARG?)) = [:](A1(?a1,ARG))*funX(ARG);

#do m=100,1,-1
id funX(ARG?)^`m' = funX(ARG,`m');
#enddo

once funX(ARG1?,ARG2?) = funX(ARG1,ARG2,0);
repeat;
repeat
 id funX(ARG1?,ARG2?,ARG3?)*funX(ARG4?,ARG5?) = funX(ARG1,ARG2,ARG3+ARG5)*funX(ARG4,ARG5,0,0);
id funX(ARG1?,ARG2?,ARG3?) = funX(ARG1,ARG2,ARG3,0,0);
once funX(ARG1?,ARG2?,0,0) = funX(ARG1,ARG2,0);
id funX(ARG1?,ARG2?,0,0) = funX(ARG1,ARG2);
endrepeat;
id funX(?a,0,0) = funX(?a);
multiply funX(0);
repeat
 id funX(ARG0?)*funX(?a,ARG1?,ARG2?) = funX(ARG0+ARG1*ARG2);

#endprocedure
*================================================================================================


*--- make Wick contractions
#do nf = `LIST'
unhide prod`nf';
.sort: Wick F;

*--- zero contractions
skip prod`nf';
L prod`nf'c0 = prod`nf';
symmetrize [:];
repeat id [:](?a1)*[:](?a2) = [:](?a1,?a2);
.sort: Wick G;

hide prod`nf'c0;

*--- subsequent contractions
#do nc = 1,`$NC`nf''
symmetrize [:];
#call CountWickTerms
repeat id [:](?a1)*[:](?a2) = [:](?a1,?a2);
id [:](?a) = fun([:](?a));
#call contract *>> make Wick contraction
#call print
.sort: Wick H (nf=`nf', nc=`nc');

*#terminate

skip prod`nf';
L prod`nf'c`nc' = prod`nf';
repeat id [:](?a1)*[:](?a2) = [:](?a1,?a2);
symmetrize [:];
#call print
.sort: Wick I (nf=`nf', nc=`nc');

hide prod`nf'c`nc';
.sort: Wick J (nf=`nf', nc=`nc');

#enddo
hide prod`nf';
.sort: Wick K (nf=`nf');

#enddo
.sort: Wick L;

*--- compose pieces with different contractions
#do nf = `LIST'
L prodd`nf' = prod`nf'c0 + ... + prod`nf'c`$NC`nf'';
#enddo
$coefxxx = coeff_;
multiply sig_($coefxxx)/$coefxxx; *>> drop combinatoric factors keeping the track of the signs
id [:]=1;
#call print;
.sort: Wick M;

unhide;
.sort: Wick N;
f [:]; *>> make [:] again noncommuting
drop;
ndrop ME;
#do nf = `LIST'
once T(?a)*Ta(`nf') = prodd`nf';
#enddo
id Tn(?a) = 1;
id Tc(?a) = 1;
#call print
.sort: Wick O1;

*--- get rid of zero-anyway terms
id [:](?a$args) = [:](?a,nargs_(?a));
id [:](?a,ARG?$nop) = [:](?a);
if ( $nop != {`NFS'-`NOFF'} ) discard;

b S;
#call print
.sort: Wick O2;
*-------------------------------------------------------------------------------------------------------------

*--- sum over the internal indices
#do l=1,`LABEL'
sum jj`l', kk`l', ll`l', rr`l', ss`l', hh`l', alph`l', bet`l', gam`l', EE`l', FF`l';
#enddo

renumber;
#call print
.sort: internal sum;



